package dev.shadowsoffire.apotheosis.adventure.boss;

import com.mojang.datafixers.kinds.Applicative;
import dev.shadowsoffire.placebo.json.NBTAdapter;
import dev.shadowsoffire.placebo.codec.PlaceboCodecs;
import com.mojang.datafixers.kinds.App;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import net.minecraft.world.entity.npc.Villager;
import net.minecraft.world.entity.ai.goal.target.NearestAttackableTargetGoal;
import net.minecraft.network.chat.ComponentContents;
import net.minecraft.util.Mth;
import java.util.stream.Collector;
import net.minecraftforge.registries.ForgeRegistries;
import dev.shadowsoffire.apotheosis.ench.asm.EnchHooks;
import java.util.HashMap;
import net.minecraft.network.chat.contents.TranslatableContents;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixHelper;
import dev.shadowsoffire.apotheosis.adventure.loot.LootController;
import net.minecraft.world.item.enchantment.EnchantmentInstance;
import java.util.stream.Collectors;
import net.minecraft.world.item.enchantment.Enchantment;
import net.minecraft.world.item.enchantment.EnchantmentHelper;
import net.minecraft.world.item.ItemStack;
import java.util.Iterator;
import net.minecraft.world.effect.MobEffectInstance;
import net.minecraft.world.effect.MobEffects;
import dev.shadowsoffire.apotheosis.adventure.AdventureConfig;
import dev.shadowsoffire.apotheosis.adventure.loot.RarityRegistry;
import dev.shadowsoffire.apotheosis.Apotheosis;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.Style;
import net.minecraft.network.chat.MutableComponent;
import dev.shadowsoffire.apotheosis.adventure.loot.LootCategory;
import net.minecraft.world.entity.EquipmentSlot;
import dev.shadowsoffire.placebo.json.GearSetRegistry;
import dev.shadowsoffire.apotheosis.util.NameHelper;
import net.minecraft.world.entity.LivingEntity;
import dev.shadowsoffire.placebo.json.RandomAttributeModifier;
import dev.shadowsoffire.placebo.json.ChancedEffectInstance;
import net.minecraft.world.entity.monster.Creeper;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.level.Level;
import java.util.function.Function;
import net.minecraft.world.entity.Mob;
import net.minecraft.util.RandomSource;
import net.minecraft.core.BlockPos;
import net.minecraft.world.level.ServerLevelAccessor;
import com.google.common.base.Preconditions;
import java.util.Optional;
import dev.shadowsoffire.apotheosis.util.SupportingEntity;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.nbt.CompoundTag;
import dev.shadowsoffire.placebo.json.GearSet;
import java.util.List;
import org.jetbrains.annotations.Nullable;
import java.util.Set;
import dev.shadowsoffire.apotheosis.adventure.loot.LootRarity;
import java.util.Map;
import net.minecraft.world.entity.EntityType;
import net.minecraft.world.entity.ai.goal.Goal;
import java.util.function.Predicate;
import net.minecraft.world.phys.AABB;
import com.mojang.serialization.Codec;
import dev.shadowsoffire.apotheosis.adventure.compat.GameStagesCompat;
import dev.shadowsoffire.apotheosis.adventure.loot.RarityClamp;
import dev.shadowsoffire.placebo.reload.WeightedDynamicRegistry;
import dev.shadowsoffire.placebo.codec.CodecProvider;

public final class ApothBoss implements CodecProvider<ApothBoss>, WeightedDynamicRegistry.ILuckyWeighted, WeightedDynamicRegistry.IDimensional, RarityClamp, GameStagesCompat.IStaged
{
    public static final Codec<AABB> AABB_CODEC;
    public static final Codec<ApothBoss> CODEC;
    public static final Predicate<Goal> IS_VILLAGER_ATTACK;
    protected final int weight;
    protected final float quality;
    protected final EntityType<?> entity;
    protected final AABB size;
    protected final Map<LootRarity, BossStats> stats;
    @Nullable
    protected final Set<String> stages;
    protected final List<GearSet.SetPredicate> gearSets;
    @Nullable
    protected final CompoundTag nbt;
    protected final Set<ResourceLocation> dimensions;
    protected final LootRarity minRarity;
    protected final LootRarity maxRarity;
    @Nullable
    protected final SupportingEntity mount;
    
    public ApothBoss(final int weight, final float quality, final EntityType<?> entity, final AABB size, final Map<LootRarity, BossStats> stats, final Optional<Set<String>> stages, final List<GearSet.SetPredicate> armorSets, final Optional<CompoundTag> nbt, final Set<ResourceLocation> dimensions, final LootRarity minRarity, final LootRarity maxRarity, final Optional<SupportingEntity> mount) {
        this.weight = weight;
        this.quality = quality;
        this.entity = entity;
        this.size = size;
        this.stats = stats;
        this.stages = stages.orElse(null);
        this.gearSets = armorSets;
        this.nbt = nbt.orElse(null);
        this.dimensions = dimensions;
        this.minRarity = minRarity;
        this.maxRarity = maxRarity;
        this.mount = mount.orElse(null);
        Preconditions.checkArgument(minRarity.ordinal() <= maxRarity.ordinal(), (Object)"Min rarity must be less than or equal to max rarity.");
    }
    
    public int getWeight() {
        return this.weight;
    }
    
    public float getQuality() {
        return this.quality;
    }
    
    public LootRarity getMinRarity() {
        return this.minRarity;
    }
    
    public LootRarity getMaxRarity() {
        return this.maxRarity;
    }
    
    public AABB getSize() {
        return this.size;
    }
    
    public EntityType<?> getEntity() {
        return this.entity;
    }
    
    public Mob createBoss(final ServerLevelAccessor world, final BlockPos pos, final RandomSource random, final float luck) {
        return this.createBoss(world, pos, random, luck, null);
    }
    
    public Mob createBoss(final ServerLevelAccessor world, final BlockPos pos, final RandomSource random, final float luck, @Nullable final LootRarity rarity) {
        final CompoundTag fakeNbt = (this.nbt == null) ? new CompoundTag() : this.nbt;
        fakeNbt.m_128359_("id", EntityType.m_20613_((EntityType)this.entity).toString());
        Mob entity = (Mob)EntityType.m_20645_(fakeNbt, (Level)world.m_6018_(), (Function)Function.identity());
        if (this.nbt != null) {
            entity.m_20258_(this.nbt);
        }
        this.initBoss(random, entity, luck, rarity);
        if (this.nbt != null) {
            entity.m_7378_(this.nbt);
        }
        if (this.mount != null) {
            final Mob mountedEntity = this.mount.create((Level)world.m_6018_(), pos.m_123341_() + 0.5, (double)pos.m_123342_(), pos.m_123343_() + 0.5);
            entity.m_7998_((Entity)mountedEntity, true);
            entity = mountedEntity;
        }
        entity.m_7678_(pos.m_123341_() + 0.5, (double)pos.m_123342_(), pos.m_123343_() + 0.5, random.m_188501_() * 360.0f, 0.0f);
        return entity;
    }
    
    public void initBoss(final RandomSource rand, final Mob entity, final float luck, @Nullable LootRarity rarity) {
        if (rarity == null) {
            rarity = LootRarity.random(rand, luck, (RarityClamp)this);
        }
        rarity = this.clamp(rarity);
        final BossStats stats = this.stats.get(rarity);
        final int duration = (entity instanceof Creeper) ? 6000 : Integer.MAX_VALUE;
        for (final ChancedEffectInstance inst : stats.effects()) {
            if (rand.m_188501_() <= inst.chance()) {
                entity.m_7292_(inst.create(rand, duration));
            }
        }
        for (final RandomAttributeModifier modif : stats.modifiers()) {
            modif.apply(rand, (LivingEntity)entity);
        }
        entity.f_21345_.f_25345_.removeIf(ApothBoss.IS_VILLAGER_ATTACK);
        final String name = NameHelper.setEntityName(rand, entity);
        final GearSet set = GearSetRegistry.INSTANCE.getRandomSet(rand, luck, (List)this.gearSets);
        set.apply((LivingEntity)entity);
        boolean anyValid = false;
        for (final EquipmentSlot t : EquipmentSlot.values()) {
            final ItemStack s = entity.m_6844_(t);
            if (!s.m_41619_() && !LootCategory.forItem(s).isNone()) {
                anyValid = true;
                break;
            }
        }
        if (!anyValid) {
            throw new RuntimeException("Attempted to apply boss gear set " + GearSetRegistry.INSTANCE.getKey((CodecProvider)set) + " but it had no valid affix loot items generated.");
        }
        int guaranteed = rand.m_188503_(6);
        for (ItemStack temp = entity.m_6844_(EquipmentSlot.values()[guaranteed]); temp.m_41619_() || LootCategory.forItem(temp) == LootCategory.NONE; temp = entity.m_6844_(EquipmentSlot.values()[guaranteed])) {
            guaranteed = rand.m_188503_(6);
        }
        for (final EquipmentSlot s2 : EquipmentSlot.values()) {
            final ItemStack stack = entity.m_6844_(s2);
            if (!stack.m_41619_()) {
                if (s2.ordinal() == guaranteed) {
                    entity.m_21409_(s2, 2.0f);
                }
                if (s2.ordinal() == guaranteed) {
                    entity.m_8061_(s2, modifyBossItem(stack, rand, name, luck, rarity, stats));
                    entity.m_6593_((Component)((MutableComponent)entity.m_7770_()).m_130948_(Style.f_131099_.m_131148_(rarity.getColor())));
                }
                else if (rand.m_188501_() < stats.enchantChance()) {
                    enchantBossItem(rand, stack, Apotheosis.enableEnch ? stats.enchLevels()[0] : stats.enchLevels()[1], true);
                    entity.m_8061_(s2, stack);
                }
            }
        }
        entity.getPersistentData().m_128379_("apoth.boss", true);
        entity.getPersistentData().m_128359_("apoth.rarity", RarityRegistry.INSTANCE.getKey((CodecProvider)rarity).toString());
        entity.m_21153_(entity.m_21233_());
        if (AdventureConfig.bossGlowOnSpawn) {
            entity.m_7292_(new MobEffectInstance(MobEffects.f_19619_, 3600));
        }
    }
    
    public static void enchantBossItem(final RandomSource rand, final ItemStack stack, final int level, final boolean treasure) {
        final List<EnchantmentInstance> ench = EnchantmentHelper.m_220297_(rand, stack, level, treasure);
        final Map<Enchantment, Integer> map = ench.stream().filter(d -> !d.f_44947_.m_6589_()).collect(Collectors.toMap(d -> d.f_44947_, d -> d.f_44948_, Math::max));
        map.putAll(EnchantmentHelper.m_44831_(stack));
        EnchantmentHelper.m_44865_((Map)map, stack);
    }
    
    public static ItemStack modifyBossItem(ItemStack stack, final RandomSource rand, final String bossName, final float luck, final LootRarity rarity, final BossStats stats) {
        enchantBossItem(rand, stack, Apotheosis.enableEnch ? stats.enchLevels()[2] : stats.enchLevels()[3], true);
        NameHelper.setItemName(rand, stack);
        stack = LootController.createLootItem(stack, LootCategory.forItem(stack), rarity, rand);
        final String bossOwnerName = String.format(NameHelper.ownershipFormat, bossName);
        final Component name = AffixHelper.getName(stack);
        if (!bossName.isEmpty()) {
            final ComponentContents 214077_ = name.m_214077_();
            if (214077_ instanceof final TranslatableContents tc) {
                final String oldKey = tc.m_237508_();
                final String newKey = "misc.apotheosis.affix_name.two".equals(oldKey) ? "misc.apotheosis.affix_name.three" : "misc.apotheosis.affix_name.four";
                final Object[] newArgs = new Object[tc.m_237523_().length + 1];
                newArgs[0] = bossOwnerName;
                for (int i = 1; i < newArgs.length; ++i) {
                    newArgs[i] = tc.m_237523_()[i - 1];
                }
                final Component copy = (Component)Component.m_237110_(newKey, newArgs).m_130948_(name.m_7383_().m_131155_(Boolean.valueOf(false)));
                AffixHelper.setName(stack, copy);
            }
        }
        final Map<Enchantment, Integer> enchMap = new HashMap<Enchantment, Integer>();
        final Iterator iterator = EnchantmentHelper.m_44831_(stack).entrySet().iterator();
        Map.Entry<Enchantment, Integer> e = null;
        while (iterator.hasNext()) {
            e = (Map.Entry<Enchantment, Integer>)iterator.next();
            if (e.getKey() != null) {
                enchMap.put(e.getKey(), Math.min(EnchHooks.getMaxLevel((Enchantment)e.getKey()), e.getValue() + rand.m_188503_(2)));
            }
        }
        if (AdventureConfig.curseBossItems) {
            final ItemStack stk = stack;
            final List<Enchantment> curses = ForgeRegistries.ENCHANTMENTS.getValues().stream().filter(e -> e.canApplyAtEnchantingTable(stk) && e.m_6589_()).collect((Collector<? super Object, ?, List<Enchantment>>)Collectors.toList());
            if (!curses.isEmpty()) {
                final Enchantment curse = curses.get(rand.m_188503_(curses.size()));
                enchMap.put(curse, Mth.m_216271_(rand, 1, EnchHooks.getMaxLevel(curse)));
            }
        }
        EnchantmentHelper.m_44865_((Map)enchMap, stack);
        stack.m_41783_().m_128379_("apoth_boss", true);
        return stack;
    }
    
    public ApothBoss validate(final ResourceLocation key) {
        Preconditions.checkArgument(this.weight >= 0, "Boss Item " + key + " has a negative weight!");
        Preconditions.checkArgument(this.quality >= 0.0f, "Boss Item " + key + " has a negative quality!");
        Preconditions.checkNotNull(this.entity, "Boss Item " + key + " has null entity type!");
        Preconditions.checkNotNull(this.size, "Boss Item " + key + " has no size!");
        if (this.minRarity != null) {
            Preconditions.checkArgument(this.maxRarity == null || this.maxRarity.isAtLeast(this.minRarity));
        }
        if (this.maxRarity != null) {
            Preconditions.checkArgument(this.minRarity == null || this.maxRarity.isAtLeast(this.minRarity));
        }
        if (this.mount != null) {
            Preconditions.checkNotNull(this.mount.entity, "Boss Item " + key + " has an invalid mount");
        }
        for (LootRarity r = this.minRarity; r != this.maxRarity; r = r.next()) {
            Preconditions.checkNotNull(this.stats.get(r));
        }
        return this;
    }
    
    public Set<ResourceLocation> getDimensions() {
        return this.dimensions;
    }
    
    public Set<String> getStages() {
        return this.stages;
    }
    
    public Codec<? extends ApothBoss> getCodec() {
        return ApothBoss.CODEC;
    }
    
    static {
        AABB_CODEC = RecordCodecBuilder.create(inst -> inst.group((App)Codec.DOUBLE.fieldOf("width").forGetter(a -> Math.abs(a.f_82291_ - a.f_82288_)), (App)Codec.DOUBLE.fieldOf("height").forGetter(a -> Math.abs(a.f_82292_ - a.f_82289_))).apply((Applicative)inst, (width, height) -> new AABB(0.0, 0.0, 0.0, (double)width, (double)height, (double)width)));
        CODEC = RecordCodecBuilder.create(inst -> inst.group((App)Codec.intRange(0, Integer.MAX_VALUE).fieldOf("weight").forGetter((Function)WeightedDynamicRegistry.ILuckyWeighted::getWeight), (App)PlaceboCodecs.nullableField(Codec.floatRange(0.0f, Float.MAX_VALUE), "quality", (Object)0.0f).forGetter((Function)WeightedDynamicRegistry.ILuckyWeighted::getQuality), (App)ForgeRegistries.ENTITY_TYPES.getCodec().fieldOf("entity").forGetter(a -> a.entity), (App)ApothBoss.AABB_CODEC.fieldOf("size").forGetter(a -> a.size), (App)LootRarity.mapCodec((Codec)BossStats.CODEC).fieldOf("stats").forGetter(a -> a.stats), (App)PlaceboCodecs.nullableField(PlaceboCodecs.setOf((Codec)Codec.STRING), "stages").forGetter(a -> Optional.ofNullable(a.stages)), (App)GearSet.SetPredicate.CODEC.listOf().fieldOf("valid_gear_sets").forGetter(a -> a.gearSets), (App)PlaceboCodecs.nullableField(NBTAdapter.EITHER_CODEC, "nbt").forGetter(a -> Optional.ofNullable(a.nbt)), (App)PlaceboCodecs.setOf(ResourceLocation.f_135803_).fieldOf("dimensions").forGetter(a -> a.dimensions), (App)LootRarity.CODEC.fieldOf("min_rarity").forGetter(a -> a.minRarity), (App)LootRarity.CODEC.fieldOf("max_rarity").forGetter(a -> a.maxRarity), (App)PlaceboCodecs.nullableField(SupportingEntity.CODEC, "mount").forGetter(a -> Optional.ofNullable(a.mount))).apply((Applicative)inst, ApothBoss::new));
        IS_VILLAGER_ATTACK = (a -> a instanceof NearestAttackableTargetGoal && ((NearestAttackableTargetGoal)a).f_26048_ == Villager.class);
    }
}
